import asyncio, requests_html_playwright, more_itertools, datetime, time, sys
from boldigger3.id_engine import parse_fasta
from bs4 import BeautifulSoup as BSoup
from requests import Response
from tqdm.asyncio import tqdm
import pandas as pd
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from requests.exceptions import RetryError
from requests.exceptions import ChunkedEncodingError, ConnectionError


# function to collect the process ids from the hdf storage
def collect_process_ids(hdf_name_results: str) -> list:
    """Function to collect all process ids from the downloaded data.

    Args:
        hdf_name_results (str): Path to the hdf storage that's generated by the data download

    Returns:
        tuple: Returns a list of unique process ids.
    """

    # read the process ids from hdf storage
    unsorted_results = pd.read_hdf(
        hdf_name_results,
        key="results_unsorted",
        columns=["process_id"],
        where="process_id != ''",
        iterator=True,
        chunksize=1000000,
    )

    # define the unique process ids variable
    unique_process_ids = set()

    # remove duplicates and empty strings from process ids
    for chunk in unsorted_results:
        unique_process_ids = unique_process_ids.union(
            set(chunk["process_id"].to_list())
        )

    # return the ids
    return unique_process_ids


# function to check if the additional data or parts of it have already been downloaded
def check_already_downloaded(hdf_name_results: str, unique_process_ids: list) -> list:
    """Function that checks if the requested download has already been performed

    Args:
        hdf_name_results (str): Path to the hdf data storage.
        unique_process_ids (list): List of unique process ids that are requested for data download.

    Returns:
        list: Returns a list of the unique process ids that have not been downloaded yet.
    """
    try:
        additional_data = pd.read_hdf(
            hdf_name_results,
            key="additional_data",
            columns=["process_id"],
            iterator=True,
            chunksize=1000000,
        )

        # collect the already downloaded ids
        downloaded_ids = set()

        # loop over all chunks of downloaded ids
        for chunk in additional_data:
            downloaded_ids = downloaded_ids.union(set(chunk["process_id"].to_list()))

        # filter unique ids
        unique_process_ids = [
            unique_id
            for unique_id in unique_process_ids
            if unique_id not in downloaded_ids
        ]

        return unique_process_ids
    except KeyError:
        return unique_process_ids


# funtion to parse a record page
def parse_record_page(html_response: object, url: str) -> list:
    """This function parses a page for any given record on bold

    Args:
        html_response (object): Requests response object to parse
        url (string): url as string to parse the process id

    Returns:
        list: A line of data representing the record data
    """
    # extract the process id from the url
    process_id = url.split("/")[-1]

    # get the status from the status code, all retries exceeded will be routed to 500
    status_code_to_status = {
        200: "public",
        404: "private",
        500: "unavailable",
    }
    status = status_code_to_status[html_response.status_code]

    # testing
    if html_response.status_code == 200:
        # transform html into beautifulsoup to parse out the data
        soup = BSoup(html_response.text, "html5lib")

        # find the specimen table
        specimen = soup.find("h1", string="Specimen").find_next("table")
        # extract the sex and lifestage
        sex = specimen.find("th", string="Sex:").find_next("td").text
        lifestage = specimen.find("th", string="Life Stage:").find_next("td").text
        # extract institution storing
        identifiers = soup.find("h1", string="Identifiers").find_next("table")
        institution_storing = (
            identifiers.find("th", string="Deposited In:").find_next("td").text
        )
        # extract the country
        collection = soup.find("h1", string="Collection").find_next("table")
        country_ocean = (
            collection.find("th", string="Country/Ocean:").find_next("td").text
        )
        # extract identifier
        attribution = soup.find("h1", string="Attribution").find_next("table")
        identifier = (
            attribution.find("th", string="Specimen Identification:")
            .find_next("td")
            .text
        )
        # extract identification method
        taxonomy = soup.find("h1", string="Taxonomy").find_next("table")
        id_method = (
            taxonomy.find("th", string="Identification Method:").find_next("td").text
        )
        # record page
        record_page = "https://portal.boldsystems.org/record/{}".format(process_id)
        # return the resulting line
        return [
            process_id,
            status,
            sex,
            lifestage,
            institution_storing,
            country_ocean,
            identifier,
            id_method,
            record_page,
        ]
    else:
        return [process_id, status] + [""] * 7


# async function to perform the request
async def as_request(url, as_session) -> list:
    """Function to perform async requests and parse the data.

    Args:
        url (_type_): url to request
        as_session (_type_): async session to perform the request with

    Returns:
        list: Returns the parsed response as list.
    """
    # request the url from BOLD
    try:
        response = await as_session.get(url, timeout=3)
    except RetryError:
        response = Response()
        response.status_code = 500
    except (ChunkedEncodingError, ConnectionError):
        response = Response()
        response.status_code = 500

    # parse the response here
    response = parse_record_page(response, url)

    # return the response to the caller to append it to hdf
    return response


# function to limit concurrency of the code
async def limit_concurrency(url, as_session, semaphore) -> object:
    """Function to limit the concurrency of requests.

    Args:
        url (_type_): url to request.
        as_session (_type_): session to use for the requests
        semaphore (_type_): semaphore object

    Returns:
        function: as request with semaphore added
    """
    async with semaphore:
        return await as_request(url, as_session)


# function to launch the async session including retry strategy
async def as_session(download_urls, semaphore) -> list:
    """Function to launch the async session and perform the data download.

    Args:
        download_urls (_type_): All urls to download.
        semaphore (_type_): semaphore to limit the concurrency

    Returns:
        list: returns a list of list with the gathered responses of the batch downloaded
    """
    # create the session with correct headers and a retry strategy
    as_session = requests_html_playwright.AsyncHTMLSession()
    as_session.headers.update(
        {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.82 Safari/537.36"
        }
    )

    # only retry if the page failed to load
    retry_strategy = Retry(
        total=5,
        status_forcelist=[500, 502, 504],
        backoff_factor=2,
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    as_session.mount("https://", adapter)

    tasks = (limit_concurrency(url, as_session, semaphore) for url in download_urls)

    return await tqdm.gather(*tasks, desc="Downloading additional data")


# function to transform the response to a pandas dataframe
def response_to_dataframe(parsed_response: list) -> object:
    """Function that transforms a list of rows to dataframe for easy saving

    Args:
        parsed_response (list): List of lists that is  returned from parsing BOLD

    Returns:
        object: Returns a pandas dataframe that can be easily saved in hdf format.
    """
    # generate dataframe
    parsed_response = pd.DataFrame(
        parsed_response,
        columns=[
            "process_id",
            "status",
            "sex",
            "lifestage",
            "institution_storing",
            "country_or_ocean",
            "identifier",
            "id_method",
            "record_page",
        ],
    )

    return parsed_response


# function to add the data to the hdf storage
def add_to_hdf(hdf_name_results: str, response: object) -> None:
    """Function to add the downloaded additional data to the hdf storage.

    Args:
        hdf_name_results (str): Path to the hdf storage.
        response (object): Dataframe to be appended to the hdf storage.
    """
    # define the item sizes for the hdf storage
    item_sizes = {
        "process_id": 30,
        "status": 11,
        "sex": 8,
        "lifestage": 80,
        "institution_storing": 150,
        "country_or_ocean": 80,
        "identifier": 80,
        "id_method": 400,
        "record_page": 70,
    }

    # save results in hdf storage
    with pd.HDFStore(
        hdf_name_results, mode="a", complib="blosc:blosclz", complevel=9
    ) as hdf_output:
        hdf_output.append(
            value=response,
            key="additional_data",
            format="t",
            data_columns=True,
            min_itemsize=item_sizes,
            complib="blosc:blosclz",
            complevel=9,
        )


# main functio to run the additional data download
def main(fasta_path: str) -> None:
    """Main function to run the additional data download. Downloads additional data
    for all public process ids, saves the to hdf first and finally add them to the
    results.

    Args:
        hdf_name_results (str): Path to the fasta file to be identified
    """
    # user output
    tqdm.write(
        "{}: Collecting process ids.".format(
            datetime.datetime.now().strftime("%H:%M:%S")
        )
    )

    # read the input fasta
    fasta_dict, fasta_name, project_directory = parse_fasta(fasta_path)

    # generate a new for the hdf storage to store the downloaded data
    hdf_name_results = project_directory.joinpath(
        "{}_result_storage.h5.lz".format(fasta_name)
    )

    # collect all process ids first
    unique_process_ids = collect_process_ids(hdf_name_results)

    # check already downloaded data
    unique_process_ids = check_already_downloaded(hdf_name_results, unique_process_ids)

    # calculate the number of batches
    batch_count = len(list(more_itertools.chunked(unique_process_ids, 10000)))
    batch_counter = 1

    # only start the download loop if there are ids to downlaod
    if unique_process_ids:
        # user output
        tqdm.write(
            "{}: Divided {} unique process ids into {} batch(es) for download.".format(
                datetime.datetime.now().strftime("%H:%M:%S"),
                len(unique_process_ids),
                batch_count,
            )
        )

        urls = [
            "https://portal.boldsystems.org/record/{}".format(idx)
            for idx in unique_process_ids
        ]

        for chunk in more_itertools.chunked(urls, 10000):
            tqdm.write(
                "{}: Downloading batch {} of {}.".format(
                    datetime.datetime.now().strftime("%H:%M:%S"),
                    batch_counter,
                    batch_count,
                )
            )
            # requests the additional data asynchronously in batches of 10000 urls
            # limit the concurrent requests to 50
            semaphore = asyncio.Semaphore(50)

            # gather the responses
            response = asyncio.run(as_session(chunk, semaphore))

            # transform to dataframe for saving
            response = response_to_dataframe(response)

            # saving the data
            add_to_hdf(hdf_name_results, response)

            # user output
            tqdm.write(
                "{}: Saving batch {} of {}.".format(
                    datetime.datetime.now().strftime("%H:%M:%S"),
                    batch_counter,
                    batch_count,
                )
            )

            # increase batch counter for user output
            batch_counter += 1
    else:
        # user output
        tqdm.write(
            "{}: Additional data has already been completly downloaded.".format(
                datetime.datetime.now().strftime("%H:%M:%S"),
                batch_counter,
                batch_count,
            )
        )



import argparse, sys, datetime, time, luddite
from boldigger3 import id_engine, additional_data_download, select_top_hit
from importlib.metadata import version


# main function to program the commandline interface
def main() -> None:
    """Function to define the commandline interface."""
    # initialize the default behaviour if boldigger3 is called without any argument
    formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=35)

    # define the parser
    parser = argparse.ArgumentParser(
        prog="boldigger3",
        description="A Python package to identify and organise sequences with the Barcode of Life Data systems.",
        formatter_class=formatter,
    )

    # display help when no argument is called
    parser.set_defaults(func=lambda x: parser.print_help())

    # add the subparsers
    subparsers = parser.add_subparsers(dest="function")

    # add the identify parser
    parser_identify = subparsers.add_parser(
        "identify", help="Run the BOLD v5 identification engine"
    )

    # add the fasta path argument
    parser_identify.add_argument(
        "fasta_file",
        help="Path to the fasta file or fasta file in the current working directory to be identified.",
        type=str,
    )

    # add the database argument
    parser_identify.add_argument(
        "--db",
        required=True,
        help="Integer that defines which database to use (1 to 8). See readme for details",
        type=int,
        choices=range(1, 9),
    )

    # add the operating mode argument
    parser_identify.add_argument(
        "--mode",
        required=True,
        help="Integer that defines which operating mode to use (1 to 3). See readme for details.",
        type=int,
        choices=range(1, 4),
    )

    # add the optional argument thresholds
    parser_identify.add_argument(
        "--thresholds",
        nargs="+",
        type=int,
        help="Thresholds to use for the selection of the top hit.",
    )

    # add version control
    # get the installed version
    current_version = version("boldigger3")
    latest_version = luddite.get_version_pypi("boldigger3")

    # give a user warning if the latest version is not installed
    if current_version != latest_version:
        print(
            "{}: Your boldigger3 version is outdated. Consider updating to the latest version.".format(
                datetime.datetime.now().strftime("%H:%M:%S")
            )
        )

    # add the version argument
    parser.add_argument("--version", action="version", version=version("boldigger3"))

    # parse the arguments
    arguments = parser.parse_args()

    # print help if no argument is provided
    if len(sys.argv) == 1:
        arguments.func(arguments)
        sys.exit()

    # only use the threshold provided by the user replace the rest with defaults
    default_thresholds = [97, 95, 90, 85]
    thresholds = []

    for i in range(4):
        try:
            thresholds.append(arguments.thresholds[i])
        except (IndexError, TypeError):
            thresholds.append(default_thresholds[i])

    if arguments.thresholds:
        # give user output
        print(
            "{}: Default thresholds changed!\n{}: Species: {}, Genus: {}, Family: {}, Order: {}".format(
                datetime.datetime.now().strftime("%H:%M:%S"),
                datetime.datetime.now().strftime("%H:%M:%S"),
                *thresholds
            )
        )

    # run the identification engine
    if arguments.function == "identify":
        # run the id engine
        id_engine.main(
            arguments.fasta_file,
            database=arguments.db,
            operating_mode=arguments.mode,
        )
        # download the additional data
        additional_data_download.main(arguments.fasta_file)
        # select the top hit
        select_top_hit.main(arguments.fasta_file, thresholds=thresholds)


# run only if called as a top level script
if __name__ == "__main__":
    main()



import datetime, more_itertools, sys
import pandas as pd
import numpy as np
from tqdm import tqdm
from boldigger3.id_engine import parse_fasta
from pathlib import Path
from tqdm import tqdm
from string import punctuation, digits


# function to sort the data by input order
def sort_by_input_order(fasta_dict: dict, dataset_df: object) -> object:
    """Function to sort a dataset or a chunk of a dataset by input order (order in the fasta file).

    Args:
        fasta_dict (dict): The fasta dict containing the input order.
        dataset_df (object): Dataframe already holding the added additional data

    Returns:
        object: The dataframe sorted by input order
    """
    # build a sorter to sort the the complete dataset by input order
    sorter = {name: idx for idx, name in enumerate(fasta_dict.keys())}

    # add the sorter column to the complete dataset
    dataset_df["sorter"] = dataset_df["id"].map(sorter)

    # name the index for sorting
    dataset_df.index.name = "index"

    # perform the sorting, remove the sorter, reset the index
    dataset_df = (
        dataset_df.sort_values(by=["sorter", "index"], ascending=[True, True])
        .drop(labels=["sorter"], axis=1)
        .reset_index(drop=True)
    )

    # return the sorted dataset
    return dataset_df


# function to clean the dataset (remove number, special chars)
def clean_dataset(dataset_df: object) -> object:
    """Funtion to clean the a chunk of the downloaded dataset. Removes names with special characters and numbers

    Args:
        dataset (object): The complete dataset as a dataframe.

    Returns:
        object: The cleaned dataset as a dataframe
    """
    dataset_df = dataset_df.copy()

    # remove punctuation and numbers from the taxonomy
    # exclude "-" since it can be part of a species name in rare cases
    # also retains no-matches
    specials = "".join([char for char in punctuation + digits if char != "-"])
    levels = ["Phylum", "Class", "Order", "Family", "Genus", "Species"]

    # clean the dataset
    for level in levels:
        dataset_df[level] = np.where(
            dataset_df[level].str.contains("[{}]".format(specials)),
            np.nan,
            dataset_df[level],
        )

    # return the cleaned dataset
    return dataset_df


# function to combine additional data and hits and sort by input order
def combine_and_sort(
    hdf_name_results: str, fasta_dict: dict, fasta_name: str, project_path: str
) -> object:
    """Combines additional data and hits, sorts the data by fasta input order
    returns the complete data as dataframe.

    Args:
        hdf_name_results (str): Path to the hdf storage.
        fasta_dict (dict): The fasta dict containing the input order
        fasta_name (str): The name of the fasta file that has to be identified
        project_path (str): Path to the project boldigger3 is working in.

    Returns:
        object: Dataframe with combined data.
    """
    # chunk to fasta dict in blocks of 10.000 seqs --> results in a maximum of 1.000.000 lines reads
    sequence_ids = more_itertools.chunked(fasta_dict.keys(), 10000)

    # loop over the chunks of ids and retrieve them from the hdf store
    for id_chunk in sequence_ids:
        # collect the chunks of the unsorted results here
        unsorted_results = []

        # create an iterator over the hdf store to not load it into memory
        hdf_iterator = pd.read_hdf(
            hdf_name_results,
            key="results_unsorted",
            where=f"id in {tuple(id_chunk)}",
            chunksize=1000000,
        )

        # collect the unsorted results from the iterator
        for chunk in hdf_iterator:
            unsorted_results.append(chunk)

        # prepare the unsorted results for this chunk of sequence ids
        unsorted_results = pd.concat(unsorted_results, ignore_index=True).reset_index(
            drop=True
        )

        # extract the process ids to collect from the additional data
        process_ids_to_retrieve = unsorted_results["process_id"]
        process_ids_to_retrieve = process_ids_to_retrieve[process_ids_to_retrieve != ""]

        # extract the additional data from the hdf based on the ids to retrieve
        additional_data = pd.read_hdf(
            hdf_name_results,
            key="additional_data",
            where=f"process_id in {process_ids_to_retrieve.to_list()}",
        )

        # transform additional data to dict, retain the column names
        additional_data = additional_data.to_dict("tight")
        column_names = additional_data["columns"][1:]
        additional_data = additional_data["data"]

        # parse the additional data into a dict in the form of process_id : [data fields] to rebuild the dataframe
        additional_data = ((record[0], record[1:]) for record in additional_data)
        additional_data = {record: data for record, data in additional_data}

        # create a dataframe with the additional data and duplicate values since each
        # process id only has to be requested once
        additional_data = pd.DataFrame(
            data=[additional_data[record] for record in process_ids_to_retrieve],
            columns=column_names,
            index=process_ids_to_retrieve.index,
        )

        # merge the unsorted results and the additional data
        complete_dataset = pd.concat([unsorted_results, additional_data], axis=1)

        # sort the complete dataset chunk by input order
        complete_dataset = sort_by_input_order(fasta_dict, complete_dataset)

        # clean the complete dataset chunk
        complete_dataset = clean_dataset(complete_dataset)

        # preset the itemsizes
        item_sizes = {
            "id": 100,
            "Phylum": 80,
            "Class": 80,
            "Order": 80,
            "Family": 80,
            "Genus": 80,
            "Species": 80,
            "bin_uri": 25,
            "request_date": 30,
            "database": 5,
            "operating_mode": 5,
            "process_id": 30,
            "status": 11,
            "sex": 8,
            "lifestage": 80,
            "institution_storing": 150,
            "country_or_ocean": 80,
            "identifier": 80,
            "id_method": 400,
            "record_page": 70,
        }

        # append each chunk to the hdf store
        with pd.HDFStore(
            hdf_name_results, mode="a", complib="blosc:blosclz", complevel=9
        ) as hdf_output:
            hdf_output.append(
                key="complete_dataset",
                value=complete_dataset,
                format="t",
                data_columns=True,
                min_itemsize=item_sizes,
                complib="blosc:blosclz",
                complevel=9,
            )

    # # save the complete dataset to excel in chunks
    dataset_for_excel = pd.read_hdf(
        hdf_name_results, key="complete_dataset", iterator=True, chunksize=1000000
    )

    # disable url writintg to supress user warning
    for idx, chunk in enumerate(dataset_for_excel, start=1):
        savename = f"{fasta_name}_bold_results_part_{idx}.xlsx"
        chunk.to_excel(project_path.joinpath(savename), engine="openpyxl", index=False)


# accepts a dataframe for any individual id
# returns the threshold to filter for and a taxonomic level
def get_threshold(hit_for_id: object, thresholds: list) -> object:
    """Function to find a threshold for a given id from the complete dataset.

    Args:
        hit_for_id (object): The hits for the respective id as dataframe.
        thresholds (list): Lists of thresholds to to use for the selection of the top hit.

    Returns:
        object: Single line dataframe containing the top hit
    """
    # find the highest similarity value for the threshold
    threshold = hit_for_id["pct_identity"].max()

    # check for no matches first
    if hit_for_id["Species"][0] == "no-match":
        return 0, "no-match"
    else:
        # move through the taxonomy if it is no nomatch hit or broken record
        if threshold >= thresholds[0]:
            return thresholds[0], "Species"
        elif threshold >= thresholds[1]:
            return thresholds[1], "Genus"
        elif threshold >= thresholds[2]:
            return thresholds[2], "Family"
        elif threshold >= thresholds[3]:
            return thresholds[3], "Order"
        # used for default thresholds --> if no hit matches the defined threshold levels, it's also a no-match
        else:
            return 0, "no-match"


## function to move the treshold one level up if no hit is found, also return the new tax level
def move_threshold_up(threshold: int, thresholds: list) -> tuple:
    """Function to move the threshold up one taxonomic level.
    Returns a new threshold and level as a tuple.

    Args:
        threshold (int): Current threshold.
        thresholds (list): List of all thresholds.

    Returns:
        tuple: (new_threshold, thresholds)
    """
    levels = ["Species", "Genus", "Family", "Order"]

    return (
        thresholds[thresholds.index(threshold) + 1],
        levels[thresholds.index(threshold) + 1],
    )


# funtion to produce and inclomplete taxonomy hit if the
# BOLD taxonomy is not complete
def return_incomplete_taxonomy(idx: int):
    incomplete_taxonomy = {
        "id": idx,
        "Phylum": "IncompleteTaxonomy",
        "Class": "IncompleteTaxnonmy",
        "Order": "IncompleteTaxonomy",
        "Family": "IncompleteTaxonomy",
        "Genus": "IncompleteTaxonomy",
        "Species": "IncompleteTaxonomy",
        "pct_identity": 0,
        "status": np.nan,
        "records": np.nan,
        "selected_level": np.nan,
        "BIN": np.nan,
        "flags": np.nan,
    }

    incomplete_taxonomy = pd.DataFrame(data=incomplete_taxonomy, index=[0])

    return incomplete_taxonomy


# function to flag the hits
def flag_hits(
    top_hits: object, hits_for_id_above_similarity: object, top_hit: object
) -> list:
    """Funtion to add a list of flag to the hits

    Args:
        top_hits (object): Dataframe with the top hits.
        hits_for_id_above_similarity (object): Dataframe of the hits above the selected similarity.
        top_hit (object): Dataframe with the selected top hit.

    Returns:
        list: Returns a list of flags.
    """
    # predefine the flags to return
    flags = [False, False, False, False, False]

    # flag 1: Reverse BIN taxonomy
    id_method = top_hits["id_method"]

    if (
        id_method.str.contains("BOLD").all()
        or id_method.str.contains("ID").all()
        or id_method.str.contains("Tree").all()
        or id_method.str.contains("BIN").all()
    ):
        flags[0] = "1"

    # flag 2: more than one group above the selected threshold
    if len(hits_for_id_above_similarity.index) > 1:
        flags[1] = "2"

    # flag 3: all of the selected top hits are private
    if top_hits["status"].isin(["private"]).all():
        flags[2] = "3"

    # flag 4: top hit is only represented by one record
    if len(top_hits.index) == 1:
        flags[3] = "4"

    # flag 5: top hit is represented by multiple bins
    if len(top_hit["BIN"].str.split(";").item()) > 1:
        flags[4] = "5"

    flags = [i for i in flags if i]
    flags = "--".join(flags)

    return flags


# function to find the top hit for a given id
def find_top_hit(hits_for_id: object, thresholds: list) -> object:
    """Funtion to find the top hit for a given ID.

    Args:
        hits_for_id (object): Dataframe with the data for a given ID
        thresholds (list): List of thresholds to perform the top hit selection with.

    Returns:
        object: Single line dataframe with the selected top hit
    """
    # get the thrshold and taxonomic level
    threshold, level = get_threshold(hits_for_id, thresholds)

    # if a no-match is found, return the no-match directly
    if level == "no-match":
        return_value = hits_for_id.query("Species == 'no-match'").head(1)

        return_value = return_value[
            [
                "id",
                "Phylum",
                "Class",
                "Order",
                "Family",
                "Genus",
                "Species",
                "pct_identity",
                "status",
            ]
        ]
        for value in ["records", "selected_level", "BIN", "flags", "status"]:
            return_value[value] = np.nan

        return return_value

    # loop through the thresholds until a hit is found
    while True:
        # copy the hits for the respective ID to perform modifications
        hits_for_id_above_similarity = hits_for_id.copy()

        # collect the idx here, to push it into the incomplete taxonomy if needed
        idx = hits_for_id_above_similarity.head(1)["id"].item()

        with pd.option_context("future.no_silent_downcasting", True):
            hits_for_id_above_similarity = hits_for_id_above_similarity.replace(
                "", np.nan
            )

        # only select hits above the selected threshold
        hits_for_id_above_similarity = hits_for_id_above_similarity.loc[
            hits_for_id_above_similarity["pct_identity"] >= threshold
        ]

        # if no hit remains move up one level until class
        if len(hits_for_id_above_similarity.index) == 0:
            try:
                threshold, level = move_threshold_up(threshold, thresholds)
                continue
            # if there is incomplete taxonomy, boldigger3 will move through all thresholds but end up here
            # return incomplete taxonomy if that is the case
            except IndexError:
                return return_incomplete_taxonomy(idx)

        # define the levels for the groupby. care about the selector string later
        all_levels = ["Phylum", "Class", "Order", "Family", "Genus", "Species"]
        levels = all_levels[: all_levels.index(level) + 1]

        # only select interesting levels (all levels above and including the selected level)
        hits_for_id_above_similarity = hits_for_id_above_similarity[levels].copy()

        # group the hits by level and then count the appearence
        hits_for_id_above_similarity = pd.DataFrame(
            {
                "count": hits_for_id_above_similarity.groupby(
                    by=levels,
                    sort=False,
                ).size()
            }
        ).reset_index()

        # if the hits still contained np.nan values, groupby will drop them:
        # if theres nothing left after the gruoupby move up one level and continue
        if len(hits_for_id_above_similarity.index) == 0:
            try:
                threshold, level = move_threshold_up(threshold, thresholds)
                continue
            # if there is incomplete taxonomy, boldigger3 will move through all thresholds but end up here
            # return incomplete taxonomy if that is the case
            except IndexError:
                return return_incomplete_taxonomy(idx)

        # sort the hits by count
        hits_for_id_above_similarity = hits_for_id_above_similarity.sort_values(
            "count", ascending=False
        )

        # select the hit with the highest count from the dataframe
        # also return the count to display in the top hit table in the end
        top_hits, top_count = (
            hits_for_id_above_similarity.head(1),
            hits_for_id_above_similarity.head(1)["count"].item(),
        )

        # generate the selector string based on the selected level
        query_string = [
            "{} == '{}'".format(level, top_hits[level].item()) for level in levels
        ]
        query_string = " and ".join(query_string)

        # query for the top hits
        top_hits = hits_for_id.query(query_string)

        # collect the bins from the selected top hit
        if threshold == thresholds[0]:
            top_hit_bins = top_hits["bin_uri"].dropna().unique()
        else:
            top_hit_bins = []

        # select the first match from the top hits table as the top hit
        top_hit = top_hits.head(1).copy()

        # add the record count to the top hit
        top_hit["records"] = top_count

        # add the selected level
        top_hit["selected_level"] = level

        # add the BINs to the top hit
        top_hit["BIN"] = ";".join(top_hit_bins)

        # define level to remove them from low level hits
        levels = ["Class", "Order", "Family", "Genus", "Species"]

        # return species level information if similarity is high enough
        # else remove higher level information form output depending on level
        if threshold == thresholds[0]:
            break
        else:
            top_hit = top_hit.assign(
                **{k: np.nan for k in levels[levels.index(level) + 1 :]}
            )
            break

    # add flags to the hits
    top_hit["flags"] = flag_hits(top_hits, hits_for_id_above_similarity, top_hit)

    # remove all data that is not needed
    top_hit = top_hit[
        [
            "id",
            "Phylum",
            "Class",
            "Order",
            "Family",
            "Genus",
            "Species",
            "pct_identity",
            "status",
            "records",
            "selected_level",
            "BIN",
            "flags",
        ]
    ]

    # return the top hit
    return top_hit


# function to save the data to excel and parquet
def save_results(project_directory: str, fasta_name: str, all_top_hits: object) -> None:
    """Function to save the identification results as Excel and Parquet.

    Args:
        project_directory (str): Project directory to save the results to.
        fasta_name (str): Name of the fasta that has to be identified.
        all_top_hits (object): Dataframe of the selected top hits.
    """
    # generate the savenames
    savename_excel = Path(project_directory).joinpath(
        "{}_identification_result.xlsx".format(fasta_name)
    )
    savename_parquet = Path(project_directory).joinpath(
        "{}_identification_result.parquet.snappy".format(fasta_name)
    )

    # save the data
    all_top_hits.to_excel(savename_excel, index=False)
    all_top_hits.to_parquet(savename_parquet)


def gather_top_hits(
    fasta_dict: dict, hdf_name_results: str, thresholds: list
) -> object:
    """Function to collect a top hit for each id in the dataset

    Args:
        fasta_dict (dict): data of the fasta file in a dictionary
        hdf_name_results (str): name of the hdf store to read the dataset from
        thresholds (list): list of thresholds for selection of top hit

    Returns:
        object: pandas dataframe with all top hits
    """
    # chunk to fasta dict in blocks of 10.000 seqs --> results in a maximum of 1.000.000 lines reads
    id_chunk_count = more_itertools.ilen(
        more_itertools.chunked(fasta_dict.keys(), 10000)
    )
    sequence_ids = more_itertools.chunked(fasta_dict.keys(), 10000)

    # gather all top hits here
    all_top_hits = []

    # loop over the chunks of ids and retrieve them from the hdf store
    for id_chunk in tqdm(
        sequence_ids,
        total=id_chunk_count,
        desc=f"Calculating top hits for {id_chunk_count} batches",
    ):
        complete_dataset_clean = pd.read_hdf(
            hdf_name_results, key="complete_dataset", where=f"id in {id_chunk}"
        )

        # calculate the top hits in chunks
        for idx in tqdm(complete_dataset_clean["id"].unique(), desc="Processing"):
            hits_for_id = (
                complete_dataset_clean.loc[complete_dataset_clean["id"] == idx]
                .copy()
                .reset_index(drop=True)
                .sort_values(by=["pct_identity"], axis=0, ascending=False)
            )

            all_top_hits.append(find_top_hit(hits_for_id, thresholds))

    all_top_hits = pd.concat(all_top_hits, axis=0).reset_index(drop=True)

    return all_top_hits


# main function to run the data sorting and top hit selection
def main(fasta_path: str, thresholds: list) -> None:
    """Main function to run data sorting and top hit selection of the downloaded data from BOLD.

    Args:
        fasta_path (str): Path to the fasta file that is identified.
        thresholds (list): Thresholds for the different taxonomic levels.
    """
    # user output
    tqdm.write(
        "{}: Loading the data for top hit selection.".format(
            datetime.datetime.now().strftime("%H:%M:%S")
        )
    )

    # read the input fasta
    fasta_dict, fasta_name, project_directory = parse_fasta(fasta_path)

    # generate a new for the hdf storage to store the downloaded data
    hdf_name_results = project_directory.joinpath(
        "{}_result_storage.h5.lz".format(fasta_name)
    )

    # user output
    tqdm.write(
        "{}: Combining results with additional data and sort by input order.".format(
            datetime.datetime.now().strftime("%H:%M:%S")
        )
    )

    # remove the complete dataset key before starting: if boldigger crashes while sorting, the sorting will be repeated
    # failsafe against incomplete sorting / combining actions
    # in the first try there is no complete dataset key, so nothing has to be done
    try:
        with pd.HDFStore(hdf_name_results) as store:
            store.remove("/complete_dataset")
    except KeyError:
        pass

    # combine downloaded data and additional data
    combine_and_sort(hdf_name_results, fasta_dict, fasta_name, project_directory)

    # user output
    tqdm.write(
        "{}: Calculating top hits.".format(datetime.datetime.now().strftime("%H:%M:%S"))
    )

    # collect the top hits
    all_top_hits = gather_top_hits(fasta_dict, hdf_name_results, thresholds)

    # Save top hits in parquet and excel
    tqdm.write(
        "{}: Saving results. This may take a while.".format(
            datetime.datetime.now().strftime("%H:%M:%S")
        )
    )

    # save the results
    save_results(project_directory, fasta_name, all_top_hits)

    # user output
    tqdm.write("{}: Finished.".format(datetime.datetime.now().strftime("%H:%M:%S")))



import datetime, sys, time, random, more_itertools, requests_html_playwright, json, gzip, pickle, os
import numpy as np
import pandas as pd
from tqdm import tqdm
from pathlib import Path
from Bio import SeqIO
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from playwright.sync_api import sync_playwright
from playwright._impl._errors import TimeoutError
from collections import OrderedDict
from boldigger3.exceptions import DownloadFinished
from json.decoder import JSONDecodeError


class BoldIdRequest:
    """A class to represent the data for a BOLD id engine request

    Attributes
    ----------

    Methods
    -------
    """

    def __init__(
        self,
        # base_url: str,
        # params: dict,
        # query_generator: object,
        # timestamp: object,
        # result_url: str,
        # queued: int,
        # processing: int,
        # completed: int,
    ):
        """Constructs the neccessary attribues for the BoldIdRequest object

        Args:
            base_url (str): Represents the base url for the post request
            params (dict): The parameters to send with the post request
            query_generator (object): A generator holding the data to send with the post request
            timestamp (object): Timestamp that is set when the request is sent to BOLD
            result_url (str): The result url to download the data from
            queued (int): Queries that are queued
            processing (int): Queries that are processing
            completed (int): Queries that are completed
        """
        self.base_url = ""
        self.params = {}
        self.query_data = []
        self.result_url = ""
        self.timestamp = None
        self.database = None
        self.operating_mode = None
        self.download_url = ""
        self.queued = ""
        self.processing = ""
        self.completed = ""


# function to read the fasta file to identify into a dictionary
def parse_fasta(fasta_path: str) -> tuple:
    """Function to read a fasta file and parse it into a dictionary.

    Args:
        fasta_path (str): Path to the fasta file to be identified.

    Returns:
        tuple: Data of the fasta file in a dict, the full path to the fasta file, the directory where this fasta file is located.
    """
    # extract the directory from the fasta path
    fasta_path = Path(fasta_path)
    fasta_name = fasta_path.stem
    project_directory = fasta_path.parent

    # use SeqIO to read the data into dict- automatically check fir the type of fasta
    fasta_dict = SeqIO.to_dict(SeqIO.parse(fasta_path, "fasta"))

    # trim header to maximum allowed chars of 99. names are preserved in the SeqRecord object
    fasta_dict = {key: value for key, value in fasta_dict.items()}

    # create a set of all valid DNA characters
    valid_chars = {
        "A",
        "C",
        "G",
        "T",
        "M",
        "R",
        "W",
        "S",
        "Y",
        "K",
        "V",
        "H",
        "D",
        "B",
        "X",
        "N",
    }

    # check all sequences for invalid characters
    raise_invalid_fasta = False

    # check if the sequences contain invalid chars
    for key in fasta_dict.keys():
        if not set(fasta_dict[key].seq.upper()).issubset(valid_chars):
            print(
                "{}: Sequence {} contains invalid characters.".format(
                    datetime.datetime.now().strftime("%H:%M:%S"), key
                )
            )
            raise_invalid_fasta = True

    if not raise_invalid_fasta:
        return fasta_dict, fasta_name, project_directory
    else:
        sys.exit()


# function to check is some of the sequences have already been downloaded
def already_downloaded(fasta_dict: dict, hdf_name_results: str) -> dict:
    """Funtion to check if any of the sequences have already been downloaded.

    Args:
        fasta_dict (dict): The dictionary with the fasta data.
        hdf_name_results (str): The savename of the hdf data storage.

    Returns:
        dict: The dictionary with the fasta data with already downloaded sequences removed.
    """
    # try to open the hdf file
    try:
        # only collect the ids from the hdf as and iterator
        idx_data = pd.read_hdf(
            hdf_name_results,
            key="results_unsorted",
            columns=["id"],
            iterator=True,
            chunksize=1000000,
        )

        # define the idx set to collect from hdf
        idx = set()

        # loop over the chunks and collect the ids
        for chunk in idx_data:
            idx = idx.union(set(chunk["id"].to_list()))

        # remove those ids from the fasta dict
        fasta_dict = {id: seq for (id, seq) in fasta_dict.items() if id not in idx}

        # return the updated fasta dict
        return fasta_dict
    except FileNotFoundError:
        # return the fasta dict unchanged
        return fasta_dict


# function to build the base urls and params
def build_url_params(database: int, operating_mode: int) -> tuple:
    """Function that generates a base URL and the params for the POST request to the ID engine.

    Args:
        database (int): Between 1 and 7 referring to the database, see readme for details.
        operating_mode (int): Between 1 and 3 referring to the operating mode, see readme for details

    Returns:
        tuple: Contains the base URL as str and the params as dict
    """

    # the database int is translated here
    idx_to_database = {
        1: "public.bin-tax-derep",
        2: "species",
        3: "all.bin-tax-derep",
        4: "DS-CANREF22",
        5: "public.plants",
        6: "public.fungi",
        7: "all.animal-alt",
        8: "DS-IUCNPUB",
    }

    # the operating mode is translated here
    idx_to_operating_mode = {
        1: {"mi": 0.94, "maxh": 25},
        2: {"mi": 0.9, "maxh": 50},
        3: {"mi": 0.85, "maxh": 100},
    }

    # params can be calculated from the database and operating mode
    params = {
        "db": idx_to_database[database],
        "mi": idx_to_operating_mode[operating_mode]["mi"],
        "mo": 100,
        "maxh": idx_to_operating_mode[operating_mode]["maxh"],
        "order": 3,
    }

    # format the base url
    base_url = "https://id.boldsystems.org/submission?db={}&mi={}&mo={}&maxh={}&order={}".format(
        params["db"], params["mi"], params["mo"], params["maxh"], params["order"]
    )

    return base_url, params


# function to build the download queue
def build_download_queue(
    fasta_dict: dict, download_queue_name: str, database: int, operating_mode: int
) -> None:
    """Function to build the download queue.

    Args:
        fasta_dict (dict): Dict that holds the data in the fasta file.
        download_queue_name (str): String that holds the path where the download queue is saved.
        database (int): Between 1 and 7 referring to the database, see readme for details.
        operating_mode (int): Between 1 and 3 referring to the operating mode, see readme for details

    """
    # initialize the download queue
    download_queue = {"waiting": OrderedDict(), "active": dict()}

    # build the base url and the params
    base_url, params = build_url_params(database, operating_mode)

    # determine the query size from the params
    query_size_dict = {0.94: 1000, 0.9: 200, 0.85: 100}
    query_size = query_size_dict[params["mi"]]

    # split the fasta dict in query sized chunks
    query_data = more_itertools.chunked(fasta_dict.keys(), query_size)

    # produce a generator that holds all sequence and key data to loop over for the post requests
    query_generators = (
        [">{}\n{}\n".format(key, fasta_dict[key].seq) for key in query_subset]
        for query_subset in query_data
    )

    for idx, query_generator in enumerate(query_generators, start=1):
        # initialize the bold id engine request
        bold_request = BoldIdRequest()
        bold_request.base_url = base_url
        bold_request.params = params
        bold_request.query_data = query_generator
        bold_request.database = database
        bold_request.operating_mode = operating_mode
        download_queue["waiting"][idx] = bold_request

    return download_queue


# function to send a post request to the BOLD id engine
def build_post_request(BoldIdRequest: object) -> object:
    """Function to send the POST request for the dataset to the BOLD id engine.

    Args:
        bold_id_request (object): A BoldIdRequest object that holds all the information needed to send the request

    Returns:
        object: Returns the BoldIdRequest object with an added result url
    """
    # send the post requests
    with requests_html_playwright.HTMLSession() as session:

        # build a retry strategy for the html session
        session.headers.update(
            {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.82 Safari/537.36"
            }
        )
        retry_strategy = Retry(total=10, backoff_factor=1)
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("https://", adapter)

        data = "".join(BoldIdRequest.query_data)

        # generate the files to send via the id engine
        files = {"file": ("submitted.fas", data, "text/plain")}

        while True:
            try:
                # submit the post request
                response = session.post(
                    BoldIdRequest.base_url, params=BoldIdRequest.params, files=files
                )

                # fetch the result
                result = json.loads(response.text)
                break
            except JSONDecodeError:
                # user output
                tqdm.write(
                    "{}: Building the request failed. Waiting 60 seconds for repeat.".format(
                        datetime.datetime.now().strftime("%H:%M:%S")
                    )
                )
                # wait 60 seconds
                time.sleep(60)

        result_url = "https://id.boldsystems.org/processing/{}".format(result["sub_id"])

        # append the resulting url
        BoldIdRequest.result_url = result_url
        BoldIdRequest.timestamp = datetime.datetime.now()

        # return the BoldIDRequest object
        return BoldIdRequest


def download_and_parse(
    BoldIdRequest: object,
    hdf_name_results: str,
    html_session: object,
) -> None:
    """This function downloads and parses the JSON from the result urls and stores it in the hdf storage

    Args:
        BoldIdRequest (object): BoldIdRequest object that holds all the data of the current request
        hdf_name_results_str (str): Name of the hdf storage to write to.
        html_session (object): session object to perform the download.
    """
    response = html_session.get(BoldIdRequest.download_url)
    response = gzip.decompress(response.content)
    content_str = response.decode("utf-8")

    # store the output dataframe here
    output_dataframe = pd.DataFrame()

    for json_record in content_str.splitlines():
        # save the results here
        json_record_results = []

        json_record = json.loads(json_record)
        # extract the sequence id first
        sequence_id = json_record["seqid"]

        # extract the results for this seq id
        results = json_record.get("results")

        # only parse if results are not empty
        if results:
            # the keys of the results are the process id|primer|bin_uri|x|x
            for key in results.keys():
                process_id, bin_uri = key.split("|")[0], key.split("|")[2]
                pident = results[key].get("pident", np.nan)
                # extract the taxonomy
                taxonomy = results.get(key).get("taxonomy", {})
                taxonomy = [
                    taxonomy.get(taxonomic_level)
                    for taxonomic_level in [
                        "phylum",
                        "class",
                        "order",
                        "family",
                        "genus",
                        "species",
                    ]
                ]

                json_record_results.append(
                    taxonomy + [pident] + [process_id] + [bin_uri]
                )
        else:
            json_record_results.append(["no-match"] * 6 + [0] + [""] + [""])

        # transform the record to dataframe to add it to the hdf storage
        json_record_results = pd.DataFrame(
            data=json_record_results,
            columns=[
                "Phylum",
                "Class",
                "Order",
                "Family",
                "Genus",
                "Species",
                "pct_identity",
                "process_id",
                "bin_uri",
            ],
        )

        # add the sequence id and the timestamp
        json_record_results.insert(0, column="id", value=sequence_id)
        json_record_results["request_date"] = pd.Timestamp.now().strftime("%Y-%m-%d %X")
        json_record_results["pct_identity"] = json_record_results[
            "pct_identity"
        ].astype("float64")

        # add the database and the operating mode
        json_record_results["database"] = BoldIdRequest.database
        json_record_results["operating_mode"] = BoldIdRequest.operating_mode

        # fill emtpy values with strings to make compatible with hdf
        json_record_results.fillna("")

        # append to the output dataframe
        output_dataframe = pd.concat([output_dataframe, json_record_results], axis=0)

    # add the results to the hdf storage
    # set size limits for the columns
    item_sizes = {
        "id": 100,
        "Phylum": 80,
        "Class": 80,
        "Order": 80,
        "Family": 80,
        "Genus": 80,
        "Species": 80,
        "process_id": 25,
        "bin_uri": 25,
        "request_date": 30,
        "database": 5,
        "operating_mode": 5,
    }

    with pd.HDFStore(
        hdf_name_results, mode="a", complib="blosc:blosclz", complevel=9
    ) as hdf_output:
        hdf_output.append(
            key="results_unsorted",
            value=output_dataframe,
            format="t",
            data_columns=True,
            min_itemsize=item_sizes,
            complib="blosc:blosclz",
            complevel=9,
        )


def download_json(
    active_queue: dict,
    hdf_name_results: str,
):
    """Function to download JSON results from the id engine and store it in hdf format

    Args:
        active_queue (dict): Queue with activebold requests
        hdf_name_results (str): name of the hdf to save the results to
        database (int): database that was queried
        operating_mode (int): operating mode that was used for the BOLD query
        request_id (int):
    """

    # start a headless playwright session to render the javascript
    # no async code needed since waiting for the rendering is required anyways
    with sync_playwright() as p:
        with requests_html_playwright.HTMLSession() as session:
            browser = p.chromium.launch(headless=True)
            page = browser.new_page()
            # loop over the active cue until any request is finished.
            while active_queue:
                for key in active_queue.keys():
                    url = active_queue[key].result_url
                    try:
                        # open the url with the browser to check if the results are fully loaded
                        page.goto(url, timeout=20000)
                    except TimeoutError:
                        continue

                    # try to find the jsonlResults selector
                    try:
                        page.wait_for_selector("#jsonlResults", timeout=20000)

                        download_url = page.query_selector(
                            "#jsonlResults"
                        ).get_attribute("href")
                        # add the download url to the BoldRequest object
                        active_queue[key].download_url = download_url
                        # parsing and download function here
                        download_and_parse(active_queue[key], hdf_name_results, session)
                        # remove the key from the dict
                        active_queue.pop(key)
                        # give user output
                        tqdm.write(
                            "{}: Request ID {} has successfully been downloaded.".format(
                                datetime.datetime.now().strftime("%H:%M:%S"), key
                            )
                        )
                        # return the active queue
                        return active_queue
                    except TimeoutError:
                        try:
                            # give user output and update if it is not found
                            queued = page.query_selector(
                                "#progress-queued"
                            ).text_content()
                            processing = page.query_selector(
                                "#progress-processing"
                            ).text_content()
                            completed = page.query_selector(
                                "#progress-completed"
                            ).text_content()
                            # add everything to the active cue object
                            active_queue[key].queued = queued
                            active_queue[key].processing = processing
                            active_queue[key].completed = completed

                            # give user output
                            tqdm.write(
                                "{}: Status of request ID {}: {}, {}, {}.".format(
                                    datetime.datetime.now().strftime("%H:%M:%S"),
                                    key,
                                    active_queue[key].queued,
                                    active_queue[key].processing,
                                    active_queue[key].completed,
                                )
                            )
                        except AttributeError:
                            continue


def main(fasta_path: str, database: int, operating_mode: int) -> None:
    """Main function to run the BOLD identification engine.

    Args:
        fasta_path (str): Path to the fasta file.
        database (int): The database to use. Can be database 1-8, see readme for details.
        operating_mode (int): The operating mode to use. Can be 1-4, see readme for details.
    """

    # user output
    tqdm.write(
        "{}: Reading input fasta.".format(datetime.datetime.now().strftime("%H:%M:%S"))
    )

    # read the input fasta
    fasta_dict, fasta_name, project_directory = parse_fasta(fasta_path)

    # generate a new for the hdf storage to store the downloaded data
    hdf_name_results = project_directory.joinpath(
        "{}_result_storage.h5.lz".format(fasta_name)
    )

    # generate a name for the download queue
    download_queue_name = project_directory.joinpath(
        "{}_download_queue.pkl".format(fasta_name)
    )

    # check if any data has been downloaded already
    fasta_dict = already_downloaded(fasta_dict, hdf_name_results)

    # if all data has already been downloaded return to stop the function
    if not fasta_dict:
        tqdm.write(
            "{}: All data has already been downloaded.".format(
                datetime.datetime.now().strftime("%H:%M:%S")
            )
        )
        return None

    try:
        with open(download_queue_name, "rb") as download_queue_file:
            download_queue = pickle.load(download_queue_file)
            # user output
            tqdm.write(
                "{}: Found unfinished downloads from previous runs. Continueing download.".format(
                    datetime.datetime.now().strftime("%H:%M:%S")
                )
            )
    except FileNotFoundError:
        # build the download queue
        tqdm.write(
            "{}: Building the download queue.".format(
                datetime.datetime.now().strftime("%H:%M:%S")
            )
        )
        download_queue = build_download_queue(
            fasta_dict, download_queue_name, database, operating_mode
        )
        with open(download_queue_name, "wb") as download_queue_file:
            pickle.dump(download_queue, download_queue_file)
        tqdm.write(
            "{}: Added {} requests to the download queue.".format(
                datetime.datetime.now().strftime("%H:%M:%S"),
                len(download_queue["waiting"]),
            )
        )

    # calculate the total amounts of downloads
    total_downloads = len(download_queue["waiting"]) + len(download_queue["active"])

    # as long as there is anything waiting or active in the download queue, keep the download running
    with tqdm(total=total_downloads, desc="Finished downloads") as pbar:
        while True:
            try:
                if download_queue["waiting"] or download_queue["active"]:
                    # as long as there are not 5 active requests in the download queue
                    # move on request from the waiting queue to the active queue
                    if len(download_queue["active"]) < 4 and download_queue["waiting"]:
                        # retrieve one request from the waiting queue
                        request_id, current_request_object = download_queue[
                            "waiting"
                        ].popitem(last=False)
                        tqdm.write(
                            "{}: Request ID {} has been moved to the active downloads.".format(
                                datetime.datetime.now().strftime("%H:%M:%S"),
                                request_id,
                            )
                        )
                        # add this request to the active queue
                        download_queue["active"][request_id] = build_post_request(
                            current_request_object
                        )
                    # check if any of the active queue objects has finished and can be saved and removed
                    else:
                        # check if any of the active downloads has been finished
                        download_queue["active"] = download_json(
                            download_queue["active"], hdf_name_results
                        )
                        # update the progress bar
                        pbar.update(1)
                    # after every iteration override the download queue as it is changes along the way
                    with open(download_queue_name, "wb") as out_stream:
                        pickle.dump(download_queue, out_stream)
                # if all downloads are finished raise download finished flag
                else:
                    raise DownloadFinished
            except DownloadFinished:
                # check if all downloads are finished: if yes: delete download queue, break the loop
                fasta_dict = already_downloaded(fasta_dict, hdf_name_results)
                # if there is any unfinished download, requeue
                if fasta_dict:
                    tqdm.write(
                        "{}: Requeuing incomplete downloads.".format(
                            datetime.datetime.now().strftime("%H:%M:%S"),
                            request_id,
                        )
                    )
                    download_queue = build_download_queue(
                        fasta_dict, download_queue_name, database, operating_mode
                    )
                    # recalculate the total downloads
                    total_downloads = len(download_queue["active"]) + len(
                        download_queue["waiting"]
                    )
                    # reset the progress bar for the second round of downloads
                    pbar.reset()
                    pbar.total(total_downloads)
                    pbar.refresh()
                else:
                    tqdm.write(
                        "{}: All downloads finished successfully.".format(
                            datetime.datetime.now().strftime("%H:%M:%S"),
                            request_id,
                        )
                    )
                    # finally remove the download queue
                    os.remove(download_queue_name)
                    break
